const { Event } = require('./schemas');
const debug = require('debug');
const log = debug('grit');

const STATUS_NEW = 'new';
const STATUS_ACQUIRED = 'acquired';
const STATUS_HANDLED = 'handled';

class EventService {
  constructor({ handlers, intentScope }, eventMachine) {
    this.handlers = handlers;
    this.intentScope = intentScope;
    this.eventMachine = eventMachine;
    let eventsToHandle = Object.keys(handlers);

    // if the intentScope is defined,
    // only handle those which are defined there.
    // an empty array means nothing will be handled..
    if (intentScope) {
      eventsToHandle = eventsToHandle.filter(handle => intentScope.indexOf(handle) > -1);
    }

    this.eventsToHandle  = eventsToHandle;
    log(`EventService to handle: ${this.eventsToHandle}`);
  }

  async getNextEvents(batchSize) {
    const currentTimestamp = new Date().getTime();

    const events = await this.getNewEventsBeforeTimestamp(
      currentTimestamp,
      batchSize
    );

    return events;
  }

  getNewEventsBeforeTimestamp(timestamp, batchSize) {
    const aggregation = [
      // get all events with status: new and any intent i can handle
      {
        $match: {
          status: { $eq: STATUS_NEW },
          intent: { $in: this.eventsToHandle }
        }
      },
      { $addFields: { timeLeft: { $subtract: ['$time', timestamp] } } },
      { $match: { timeLeft: { $lte: 0 } } },
      // highest priority first, if there are two events with priority 100, the older one should get picked, if there are two events which should have been excecuted days ago, take the one with the higher priority first
      { $sort: { priority: -1, timeLeft: 1 } },
      { $limit: batchSize }
    ];

    return Event.aggregate(aggregation);
  }

  markEventAsHandled({ _id }) {
    return Event.findOneAndUpdate(
      { _id, status: STATUS_ACQUIRED },
      { status: STATUS_HANDLED },
      { new: true }
    );
  }

  async handleEvent(event) {
    const handler = this.handlers[event.intent];
    if (!handler) {
      // TODO: can this be removed since only fetch events to which we have handlers?
      throw new Error(`No handler defined for intent '${event.intent}'`);
    } else {
      await handler(event, this.eventMachine);
    }

    const handledEvent = await this.markEventAsHandled(event);

    return handledEvent;
  }

  acquireEvent({ _id }) {
    return Event.findOneAndUpdate(
      {
        _id,
        status: STATUS_NEW
      },
      {
        status: STATUS_ACQUIRED
      },
      {
        new: true
      }
    );
  }
}

module.exports = EventService;
