const connectToDatabase = require('./database');
const backoff = require('./backoff');
const EventService = require('./EventService');
const debug = require('debug');
const log = debug('grit');
const { Event } = require('./schemas');

const DEFAULT_PRIORITY_DEFAULT = 50;

const STATUS_NEW = 'new';

class MongoDBEventmachine {
  constructor(options, handlers) {
    this.listening = false;
    this.dbUrl = options.dbUrl;
    this.batchSize = options.batchSize || 1;
    this.DEFAULT_PRIORITY =
      options.DEFAULT_PRIORITY || DEFAULT_PRIORITY_DEFAULT;

    // if there are no handlers, this is only a producer not a consumer/worker
    this.isConsumer = !!handlers;

    this.handlers = handlers || [];
    this.intentScope = options.intentScope;

    this.eventService = new EventService({ handlers: this.handlers, intentScope: this.intentScope }, this);
    this.createNewEvent = this.createNewEvent.bind(this);
    this.fetchAcquireAndHandleNextEvents = this.fetchAcquireAndHandleNextEvents.bind(
      this
    );
    this.acquireAndHandleEvent = this.acquireAndHandleEvent.bind(this);
  }

  async listen() {
    this.db = await connectToDatabase({ dbUrl: this.dbUrl }).catch(err =>
      console.error(err)
    );

    if (!this.db) {
      throw new Error(`Cannot connect to db.`);
    }

    this.listening = true;

    // if this is a consumer, poll the event queue using a backoff algorithm
    if (this.isConsumer) {
      // start a backoff series for fetching, acquiring and handling any next event
      backoff(undefined, this.fetchAcquireAndHandleNextEvents);
    }
  }

  async stopListening() {
    if (!this.db) {
      throw new Error(`Cannot close db, was not open.`);
    }

    await this.db.close();
    this.listening = false;
  }

  async acquireAndHandleEvent(nextEvent) {
    // try to acquire the event (some other worker might try that as well at the same time)
    const acquiredEvent = await this.eventService.acquireEvent(nextEvent);
    if (!acquiredEvent) {
      return false;
    }
    log({ acquiredEvent: acquiredEvent._id });

    // we now have an acquired event and need to handle it
    const handledEvent = await this.eventService.handleEvent(acquiredEvent);
    log({ handledEvent: handledEvent._id });
    return true;
  }

  async fetchAcquireAndHandleNextEvents() {
    if (!this.listening) { return { backoff: true }; }

    log('fetching events');

    // first check if there is an event that needs to be executed now or before now
    const nextEvents = await this.eventService.getNextEvents(this.batchSize);
    if (nextEvents.length === 0) {
      return { backoff: true };
    }

    const acquireAndHandleBatch = [];

    nextEvents.forEach(nextEvent => {
      log({ nextEvent: nextEvent._id });
      acquireAndHandleBatch.push(this.acquireAndHandleEvent(nextEvent));
    });

    const handledEventsBatch = await Promise.all(acquireAndHandleBatch);

    if (handledEventsBatch.filter(event => !!event).length === 0) {
      return { backoff: true };
    } else {
      return { backoff: false };
    }
  }

  createNewEvent({ intent, payload, time, priority = this.DEFAULT_PRIORITY }) {
    return Event.create({
      intent,
      payload,
      time,
      status: STATUS_NEW,
      priority
    });
  }
}

module.exports = MongoDBEventmachine;
